module JSONError
  class Base < StandardError
    DEFAULT_STATUS = 500
    NULL_TOKEN = '~'

    attr_reader :context, :key, :options

    class << self
      attr_reader :key_alias

      def inherited(base)
        base.instance_variable_set(:@translated_properties, translated_properties)
        base.instance_variable_set(:@custom_properties, custom_properties)
        base.instance_variable_set(:@key_alias, key_alias)
      end

      def base?
        self == JSONError.base
      end

      def disallowed_property_setting!(property)
        raise(Errors::DisallowedPropertySetting, property: property, class: self)
      end

      def key(key = nil, options = {})
        return @key if key.nil?
        disallowed_property_setting!(:key) if base?

        @key_alias = options[:as] if options[:as].is_a?(Symbol)
        @key ||= key
      end

      def status(status = nil)
        return @status if status.nil?
        disallowed_property_setting!(:status) if base?

        @status = status
      end

      def properties
        [*default_properties, *translated_properties, *custom_properties]
      end

      def default_properties
        %i(status key)
      end

      def translated_properties(*translated_properties)
        return @translated_properties if translated_properties.empty?
        disallowed_property_setting!(:translated_properties) if base?

        translated_properties.each do |property|
          define_method(property) { translate(property) }
        end
        @translated_properties = [*@translated_properties, *translated_properties].uniq
      end

      def custom_properties(*custom_properties)
        return @custom_properties if custom_properties.empty?
        disallowed_property_setting!(:custom_properties) if base?

        attr_accessor(*custom_properties)
        @custom_properties = [*@custom_properties, *custom_properties].uniq
      end

      def visible_properties(*visible_properties)
        return(@visible_properties || properties) if visible_properties.empty?
        disallowed_property_setting!(:visible_properties) if base?

        @visible_properties = visible_properties.uniq
      end
    end

    def initialize(attrs = {})
      @key = self.class.key || attrs.fetch(:key)
      @context = attrs[:context]
      assign_custom_properties(attrs)
      assign_options(attrs)
      super(message)
    end

    def status
      self.class.status || 500
    end

    def to_json(*args)
      visible_properties_hash.to_json(*args)
    end

    private

    def assign_custom_properties(attrs = {})
      attrs = attrs.dup.delete_if { |k, _v| !custom_properties.include?(k) }
      custom_properties.each do |property|
        instance_variable_set("@#{property}", attrs[property])
      end
    end

    def assign_options(attrs = {})
      reserved_names = %i(key context) + custom_properties
      @options = attrs.dup.delete_if { |k, _v| reserved_names.include?(k) }
      options.merge!(default: NULL_TOKEN)
    end

    def translate(node, translation_options = {})
      translation_options = options.merge(translation_options)
      translation = I18n.t("#{translation_path}.#{node}", translation_options)
      translation unless translation == NULL_TOKEN
    end

    def translation_path
      context.nil? ? "json_errors.#{key}" : "json_errors.#{context}.#{key}"
    end

    def visible_properties_hash
      h = Hash[visible_properties.map { |property| [property, send(property)] }]
      h[key_alias] = h.delete(:key) unless key_alias.nil?
      h
    end

    def visible_properties
      self.class.visible_properties
    end

    def custom_properties
      self.class.custom_properties || []
    end

    def key_alias
      self.class.key_alias
    end
  end
end
